#!/usr/bin/env python
# coding=UTF-8

import sys, os, platform, time, subprocess, stat
import re
import libxml2, libxslt
import Image
import EXIF
import pdb

import config
import html5out as htmlout

from os.path         import basename, dirname, splitext, join, exists, getmtime
from tempfile        import mkstemp
from codecs          import open
from shutil          import copy
from xml.sax         import handler, make_parser
from xml.sax.handler import feature_namespaces
from datetime        import datetime

# work around for "Suspension not allowed here" error
# see http://mail.python.org/pipermail/image-sig/1999-August/000816.html
import ImageFile
ImageFile.MAXBLOCK = 1000000 # default is 64k

# regular expressions for iso dates
date1_re = re.compile( '([0-9]{4})-([0-9]{2})-([0-9]{2})' )
date2_re = re.compile( '([0-9]{4})-([0-9]{2})' )

# regular expression for extracting date information
edate_re = re.compile( '([0-9]{4}):([0-9]{2}):([0-9]{2}) ([0-9]{2}):([0-9]{2}):([0-9]{2})' )

##############################################################################
##############################################################################
#
# local functions
#
##############################################################################
##############################################################################

#
# return operating system
#             
def get_system ():
    system = platform.system()
    if system in ( 'Microsoft', 'MSWin32', 'MSWin64' ):
        system = 'Windows'
    return system

#
# remove redundant whitespaces from a string
# 
def normalise_whitespace ( text ):
    return ' '.join( text.split() )

#
# search for PHOTOS.xml in given directory lists
#
def searchdir ( albums, dirname, names ):
    dir = basename( dirname )
    for file in names:
        if basename( file ) == 'PHOTOS.xml':
            albums.append( join( dir, file ) )

#
# convert EXIF date to datetime object
#
def exifdate_to_datetime ( exif_date ):
    m = edate_re.search( exif_date )
    if m != None:
        return datetime( int( m.group( 1 ) ), int( m.group( 2 ) ), int( m.group( 3 ) ), 
                         int( m.group( 4 ) ), int( m.group( 5 ) ), int( m.group( 6 ) ) )

#
# read date from EXIF data in given file
#
def exif_date ( photofile ) :
    f     = open( photofile, 'rb' )
    tags  = EXIF.process_file( f, stop_tag = 'DateTimeOriginal' )
    f.close()
    return exifdate_to_datetime( tags['EXIF DateTimeOriginal'].values )
    
##############################################################################
##############################################################################
#
# content handler for images, etc.
#
##############################################################################
##############################################################################

class ScanTags ( handler.ContentHandler ) :

    #
    # ctor
    #
    def __init__ ( self, album ):
        self.active = { 'album'       : False,
                        'title'       : False,
                        'date'        : False,
                        'prev'        : False,
                        'next'        : False,
                        'type'        : False,
                        'photo'       : False,
                        'video'       : False,
                        'img'         : False,
                        'file'        : False,
                        'desc'        : False,
                        'orientation' : False,
                        'format'      : False,
                        'tags'        : False }
        self.content = ''
        self.album = album
        self.entry   = None

    #
    # enter a tag
    #
    def startElement ( self, name, attrs ):
        # mark tag as active
        if not name in self.active:
            print( '  opening unknown tag "' + name + '"' )
        else:
            self.active[name] = True

        # make new photo/video entry
        if name == 'photo':
            self.entry = Photo()
        elif name == 'video':
            self.entry = Video()
        
    #
    # leave a tag
    #
    def endElement ( self, name ):
        # mark tag as inactive
        if not name in self.active:
            # self.content = self.content + '<' + name + '>' + 
            print( '  closing unknown tag "' + name + '"' )
        else:
            # remove unwanted whitespaces from tag content
            content = normalise_whitespace( self.content )

            # copy tag content to global variables or current photo
            if   self.active['title']:  self.album.title = content
            elif self.active['date']:   self.album.date  = content
            elif self.active['next']:   self.album.next  = content
            elif self.active['prev']:   self.album.prev  = content
            elif self.active['photo'] or self.active['video']:
                if   self.active['file']:        self.entry.file        = content
                elif self.active['img']:         self.entry.file        = content
                elif self.active['desc']:        self.entry.desc        = content
                elif self.active['orientation']: self.entry.orientation = content
                elif self.active['format']:      self.entry.format      = content
                elif self.active['tags']:        self.entry.tags        = content
            elif self.active['tags']:   self.album.tags = content

            # mark tag as inactive
            self.active[name] = False

            # reset tag content
            self.content = ''

            # put current photo/video to image database if photo tag was finished
            if self.entry != None and ( name == 'photo' or name == 'video' ):
                self.album.db.append( self.entry )

    #
    # called for tag content
    #
    def characters ( self, ch ):
        self.content = self.content + unicode( ch )


##############################################################################
##############################################################################
#
# class representing single photo
#
##############################################################################
##############################################################################

class Photo :

    def __init__ ( self, file = '', orientation = 'landscape', format = '' ) :
        # default settings
        self.type        = 'photo'
        self.file        = file
        self.desc        = ''
        self.orientation = orientation
        self.format      = format
        self.tags        = ''

    #################################################################
    #
    # return name of lores and thumbnail file
    #
    #################################################################

    def lores_file ( self, basedir = '' ) :
        "return name of low-resolution file associated with current photo"
        name = splitext( basename( self.file ) )[0]
        name = name + '.jpg'
        if basedir != '' :
            name = join( basedir, name )
        return name
        
    def thumb_file ( self, basedir = '' ) :
        "return name of thumbnail file associated with current photo"
        name = splitext( basename( self.file ) )[0]
        name = 'thumb_' + name + '.jpg'
        if basedir != '' :
            name = join( basedir, name )
        return name

    def content_file ( self, basedir = '' ) :
        "return name of content file associated with current photo"
        if basedir == '' :
            return splitext( basename( self.file ) )[0] + '.html'
        else :
            return join( basedir, splitext( basename( self.file ) )[0] + '.html' )
    
    #################################################################
    #
    # return creation time of original photo
    #
    #################################################################
    
    def datetime ( self ) :
        return exif_date( self.file )

    #################################################################
    #
    # generate lores and thumbnail file
    #
    #################################################################

    def gen_lores ( self, db ):

        basedir = db.directory()
        orig    = self.file
        lores   = self.lores_file( basedir )

        if config.verbosity > 1 :
            print( '    building ' + lores )

        if self.orientation == 'landscape':
            if self.format == '4x3': resolution = ( 1050, 700 )
            else:                   resolution = ( 1050, 700 )
        else:
            if self.format == '4x3': resolution = ( 675, 900 )
            else:                   resolution = ( 600, 900 )

        if get_system() == 'Linux':
        
            # on Linux, if ImageMagick is available, use 'convert' due to better results
            try:
                convert = 'convert -quality 80 -thumbnail ' + str(resolution[0]) + 'x' + str(resolution[1]) + ' ' + orig + ' ' + lores
                # file = os.popen( convert, 'r' )
                # file.close()
                process = subprocess.Popen( convert, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE )
                process.wait()
            except Exception, e:
                print( '(Photo) error calling "convert" on \"' + lores + '\" : ' + str( e ) )
                sys.exit( 1 )

            # also apply "jpegoptim" if available
            try:
                shrink  = 'jpegoptim -m80 ' + lores
                # file = os.popen( shrink, 'r' )
                # file.close()
                process = subprocess.Popen( shrink, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE )
                process.wait()
                # jpegoptim does not set group/other read permissions
                st = os.stat( lores )
                os.chmod( lores, st.st_mode | stat.S_IRGRP | stat.S_IROTH )
            except Exception, e:
                print( '(Photo) error calling "jpegoptim" on \"' + lores + '\" : ' + str( e ) )
                sys.exit( 1 )
        else:
            # in all other cases use ImageLib
            try:
                picture = Image.open( orig )
                picture.thumbnail( resolution, Image.ANTIALIAS )
                if picture.mode != 'RGB': 
                    picture = picture.convert( 'RGB' ) 
                picture.save( lores, 'JPEG', quality = 85, optimize = 1 )
            except IOError, e:
                print( '(Photo) error creating "' + lores + '" : ' + str( e ) )
                sys.exit( 1 )

    def gen_thumb ( self, db ):

        basedir = db.directory()
        lores   = self.lores_file( basedir )
        thumb   = self.thumb_file( basedir )

        if config.verbosity > 1 :
            print( '    building ' + thumb )

        if self.orientation == 'landscape':
            if self.format == '4x3': resolution = ( 178, 133 )
            else:                   resolution = ( 200, 133 )
        else:
            if self.format == '4x3': resolution = ( 100, 133 )
            else:                   resolution = (  89, 133 )

        try:
            picture = Image.open( lores )
            picture = picture.resize( resolution, Image.ANTIALIAS )
            if picture.mode != 'RGB': 
                picture = picture.convert( 'RGB' ) 
            picture.save( thumb, 'JPEG', quality = 90, optimize = 1 )
        except IOError, e:
            print( '(Photo) error creating "' + thumb + '" : ' + str( e ) )
            sys.exit( 1 )

    #################################################################
    #
    # write database entry to given file
    #
    #################################################################

    def write ( self, file ) :
        file.write( '  <photo>\n' )
        file.write( '    <file>'        + self.file        + '</file>\n' )
        file.write( '    <desc>'        + self.desc        + '</desc>\n' )
        file.write( '    <orientation>' + self.orientation + '</orientation>\n' )
        file.write( '    <format>'      + self.format      + '</format>\n' )
        file.write( '    <tags>'        + self.tags        + '</tags>\n' )
        file.write( '  </photo>\n' )
        

##############################################################################
##############################################################################
#
# class representing single video
#
##############################################################################
##############################################################################

class Video :

    def __init__ ( self, file = '', orientation = 'landscape', format = '' ) :
        # default settings
        self.type        = 'video'
        self.file        = file
        self.desc        = ''
        self.orientation = orientation
        self.format      = format
        self.tags        = ''

    #################################################################
    #
    # return name of lores and thumbnail file
    #
    #################################################################

    def lores_file ( self, basedir = '' ) :
        "return name of low-resolution file associated with current video"
        if basedir == '' :
            return config.lores( basedir, self )
        else:
            return join( basedir, config.lores( basedir, self ) )
        
    def thumb_file ( self, basedir = '' ) :
        "return name of thumbnail file associated with current video"
        name = splitext( basename( self.file ) )[0]
        name = 'thumb_' + name + '.gif'
        if basedir != '' :
            name = join( basedir, name )
        return name

    def content_file ( self, basedir = '' ) :
        "return name of content file associated with current video"
        if basedir == '' :
            return splitext( basename( self.file ) )[0] + '.html'
        else :
            return join( basedir, splitext( basename( self.file ) )[0] + '.html' )
    
    #################################################################
    #
    # return creation time of original video
    #
    #################################################################
    
    def datetime ( self ) :
        # look for corresponding photo
        basename = splitext( self.file )[0]
        for photofile in [ basename + '.jpg', basename + '.JPG' ] :
            if exists( photofile ) :
                return exif_date( photofile )

        # fall back to modification time
        return datetime.fromtimestamp( getmtime( self.file ) )

    #################################################################
    #
    # generate lores and thumbnail file
    #
    #################################################################

    def gen_lores ( self, db ):

        basedir = db.directory()
        orig    = self.file
        lores   = self.lores_file( basedir )

        if config.verbosity > 1 :
            print( '    building ' + lores )

        try:
            convert = join( config.bin_dir, 'webvideo ' + orig + ' ' + lores )
            # file = os.popen( convert, 'r' )
            # file.close()
            process = subprocess.Popen( convert, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE )
            process.wait()
        except Exception, e:
            print( '(Video) error calling "webvideo" on \"' + lores + '\" : ' + str( e ) )
            sys.exit( 1 )

    def gen_thumb ( self, db ):

        basedir = db.directory()
        orig    = self.file
        thumb   = self.thumb_file( basedir )

        if config.verbosity > 1 :
            print( '    building ' + thumb )

        try:
            convert = join( config.bin_dir, 'mkgif ' + orig + ' ' + thumb )
            # file = os.popen( convert, 'r' )
            # file.close()
            process = subprocess.Popen( convert, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE )
            process.wait()
        except Exception, e:
            print( '(Video) error calling "mkgif" on \"' + thumb + '\" : ' + str( e ) )
            sys.exit( 1 )

    #################################################################
    #
    # write database entry to given file
    #
    #################################################################

    def write ( self, file ) :
        file.write( '  <video>\n' )
        file.write( '    <file>'        + self.file        + '</file>\n' )
        file.write( '    <desc>'        + self.desc        + '</desc>\n' )
        file.write( '    <orientation>' + self.orientation + '</orientation>\n' )
        file.write( '    <format>'      + self.format      + '</format>\n' )
        file.write( '    <tags>'        + self.tags        + '</tags>\n' )
        file.write( '  </video>\n' )


##############################################################################
##############################################################################
#
# Album class managing photos of a single date
#
##############################################################################
##############################################################################

class Album :

    #################################################################
    #
    # construct album defined by content in <dbfile>
    #
    #################################################################

    def __init__ ( self, dbfile ):
        #
        # default values
        # 

        self.dbfile = dbfile
        self.mtime  = time.time() # current time by default
        self.db     = []

        self.title  = ''
        self.tags   = ''
        self.date   = ''
        self.prev   = ''
        self.next   = ''
        
        #
        # parse PHOTO database
        #

        if exists( dbfile ) :
            self.mtime  = getmtime( dbfile )
            
            if config.verbosity > 2 :
                print( '      parsing "' + dbfile + '"' )

            # set up parser
            parser = make_parser()

            parser.setFeature( feature_namespaces, 0 )

            # define content handler
            parser.setContentHandler( ScanTags( self ) )

            # parse file
            file = open( dbfile, 'r' )
            parser.parse( file )

    #################################################################
    #
    # manage internal database
    #
    #################################################################

    #
    # return entry object for given img file
    #
    def get_entry ( self, imgfile ) :
        for entry in self.db :
            if imgfile == entry.file :
                return entry
        return None

    #
    # return true of given image file is already in album
    #
    def contains ( self, imgfile ) :
        if self.get_entry( imgfile ) != None : return True
        else :                                 return False

    #
    # append new entry to database 
    # - if <replace> == True, old entries are replaced
    #   otherwise, new entry is rejected
    #
    def append ( self, new_entry, replace = False ) :
        for i, entry in enumerate( self.db ) :
            if entry.file == new_entry.file :
                if replace :
                    self.db[ i ] = new_entry
                return
        # as <new_entry> not in list, just append it
        self.db.append( new_entry )

    #
    # remove entry with given filename
    #
    def remove ( self, file, delete_files = True ) :
        for entry in self.db :
            if entry.file == file :
                # delete album files associated with entry
                if delete_files :
                    album_dir = dirname( self.dbfile )
                    for file in [ entry.lores_file( album_dir ), 
                                  entry.thumb_file( album_dir ),
                                  entry.content_file( album_dir ) ] : 
                        if exists( file ) :
                            if config.verbosity > 1 :
                                print "removing " + file
                            os.rename( file, file + config.backup_ext )
                # remove entry from DB
                self.db.remove( entry )
                return
        
    #################################################################
    #
    # return directory, index file, etc.
    #
    #################################################################

    def directory ( self, relative = False ) :
        subdir = ''
        m = date1_re.search( self.date )
        if m != None :
            year   = int(m.group(1))
            month  = int(m.group(2))
            day    = int(m.group(3))
            subdir = join( str(year), "%02d-%02d" % ( month, day ) )
        else:
            m = date2_re.search( self.date )
            if m != None :
                year   = int(m.group(1))
                month  = int(m.group(2))
                subdir = join( str(year), "%02d" % month )
            else:
                print( '(Album) unsupported date: ' + self.date )
                sys.exit( 1 )

        if relative:
            return subdir
        else:
            return join( config.top_dir, subdir )

    def index_link ( self ) :
        return join( self.directory( relative = True ), 'index.html' )

    def index_file ( self ) :
        return join( self.directory( relative = False ), 'index.html' )

    #################################################################
    #
    # build index files
    #
    #################################################################

    def make_index ( self, force = None ):
        idx_file = self.index_file()

        if ( force != None and force['html'] ) or \
                not exists( idx_file ) or         \
                self.mtime > getmtime( idx_file ):
            if config.verbosity > 0 :
                print( '  writing ' + idx_file )
            htmlout.write_day( dir, idx_file, self );

    #################################################################
    #
    # build all content files, e.g. all lores/thumbnails images and xml/html files
    #
    #################################################################

    def make_content ( self, copyimg, force = None ):
        for idx in range( 0, len(self.db) ):

            if idx > 0: prev = self.db[ idx - 1 ]
            else:       prev = None

            if idx < len(self.db)-1: next = self.db[ idx + 1 ]
            else:                    next = None

            self.make_img_content( self.db[idx], prev, next, copyimg, force )

    #################################################################
    #
    # build content for single entry, i.e. lores/thumbnail image and xml/html file
    #
    #################################################################

    def make_img_content ( self, img, prev, next, copyimg, force = None ):

        is_photo = False

        if img.type == 'photo': 
            is_photo = True

        basedir = self.directory()
        orig    = img.file
        lores   = img.lores_file( basedir )
        thumb   = img.thumb_file( basedir )

        #
        # generate lores image
        #
    
        if ( ( force != None and force['lores'] ) or \
                              not exists( lores )              or \
                              getmtime( orig ) > getmtime( lores ) ):
            img.gen_lores( self )

        #
        # generate thumbnail
        #

        if ( ( force != None and force['thumb'] ) or \
                              not exists( thumb )              or \
                              getmtime( lores ) > getmtime( thumb ) ):
            img.gen_thumb( self )

        #
        # copy original image if requested
        #

        origcopy = join( basedir, config.copy_dir, basename( orig ) )

        if copyimg and ( not exists( origcopy ) or getmtime( orig ) > getmtime( origcopy ) ):
            try:
                os.makedirs( join( basedir, config.copy_dir ) )
            except:
                pass

            try:
                copy( orig, origcopy )
            except IOError, e:
                print( '(Album) could not copy "' + orig + '" to "' + origcopy + '" : ' + str( e ) )
                sys.exit( 1 )

        #
        # generate xml/html file for image
        #

        img_file = img.content_file( basedir )

        if ( force != None and force['html'] ) or \
                not exists( img_file )         or \
                self.mtime > getmtime( img_file ):
            if config.verbosity > 1 :
                print( '    writing ' + img_file )
            htmlout.write_photo( basedir, img_file, img, self, prev, next )

    #################################################################
    #
    # write photo database to file <filename> (or self.dbfile)
    #
    #################################################################

    def write ( self, filename = None ):
        
        if filename == None :
            filename = self.dbfile

        if config.verbosity > 1 :
            print( '    writing ' + filename )

        # update filename
        self.dbfile = filename

        # create backup
        if exists( filename ) and config.backup :
            os.rename( filename, filename + config.backup_ext )

        dbfile = open( filename, encoding='utf-8', mode='w' )
        
        dbfile.write( '<album>\n' )

        # write album header
        dbfile.write( '  <title>' + self.title + '</title>\n' )
        dbfile.write( '  <tags>'  + self.tags  + '</tags>\n' )
        dbfile.write( '  <date>'  + self.date  + '</date>\n' )
        dbfile.write( '  <prev>'  + self.prev  + '</prev>\n' )
        dbfile.write( '  <next>'  + self.next  + '</next>\n' )

        # write photo/video entries
        for entry in self.db:
            entry.write( dbfile )

        dbfile.write( '</album>\n' )

        dbfile.close()

        # update modification time
        self.mtime = getmtime( filename )

    #################################################################
    #
    # sort internal database according to time of entries
    #
    #################################################################

    def sort ( self, filename = None ):
        
        def datetime ( entry ):
            return entry.datetime()

        self.db.sort( key = datetime )
        
## EOF
